--!parent
--!encoding=utf-8

--****************************************************************************
-- @file			LSW CORE LIB: BYTES
-- @author   	Valentin Schmidt
-- @version  	0.1
--****************************************************************************

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
	return me
end

----------------------------------------
-- converts text in bytearray from one encoding to another
-- requires D11.5+
-- for supported encodings check: put _system.getInstalledCharSets()
-- ["big5", "cp1026", "cp866", "ebcdic-cp-us", "gb2312", "ibm437", "ibm737", "ibm775",
-- "ibm850", "ibm852", "ibm857", "ibm861", "ibm869", "iso-8859-1", "iso-8859-15", "iso-8859-2",
-- "iso-8859-4", "iso-8859-5", "iso-8859-7", "iso-8859-9", "johab", "koi8-r", "koi8-u", 
-- "ks_c_5601-1987", "macintosh", "shift_jis", "us-ascii", "utf-16", "utf-16be", "utf-7", 
-- "utf-8", "windows-1250", "windows-1251", "windows-1252", "windows-1253", "windows-1254",
-- "windows-1255", "windows-1256", "windows-1257", "windows-1258", "windows-874", "x-ebcdic-greekmodern", 
-- "x-mac-ce", "x-mac-cyrillic", "x-mac-greek", "x-mac-icelandic", "x-mac-turkish"]
-- 
-- @param {bytearray} inputBa
-- @param {string} inputEncoding
-- @param {string} outputEncoding
-- @return {bytearray}
----------------------------------------
on convertEncoding (me, inputBa, inputEncoding, outputEncoding)
  inputBa.position = 1
  utf8 = inputBa.readRawString(inputBa.length, inputEncoding)
  tmp = bytearray()
  tmp.writeString(utf8, outputEncoding)
  tmp.position = 5
  return tmp.readByteArray(tmp.length-4)
end

----------------------------------------
-- Checks bytearray for Byte-Order-Mark (BOM), returns BOM-type or false
-- @param {bytearray} ba
-- @return {string|false} - either "UTF-8", "UTF-16", "UTF-16BE" or false
----------------------------------------
on hasBOM (me, ba)
	if ba.length>=3 then
		if (ba[1]=239 AND ba[2]=187 AND ba[3]=191) then 
			return "UTF-8"
		end if
	end if
	if ba.length>=2 then
		if (ba[1]=255 AND ba[2]=254) then return "UTF-16" -- LE
		else if (ba[1]=254 AND ba[2]=255) then return "UTF-16BE" -- BE
	end if
	return false
end

----------------------------------------
-- Checks bytearray for invalid UTF-8 sequences
-- Notice: since ASCII is a UTF-8 subset, function returns true for pure ASCII data
-- @param {bytearray} ba
-- @return {bool}
----------------------------------------
on isUtf8 (me, ba)
  len = ba.length
  repeat with i = 1 to len
    ord = ba[i]
    
    if (ord < 128) then next repeat -- 0bbbbbbb
    else if(bitAnd(ord,224)=192 AND ord>193) then n = 1 -- 110bbbbb (C0-C1)
    else if(bitAnd(ord,240)=224) then n = 2 -- 1110bbbb
    else if(bitAnd(ord,248)=240 AND ord<245) then n = 3 -- 11110bbb (F5-FF)
    else return false -- invalid UTF-8 sequence
    
    repeat with c = 1 to n
    	i = i+1
    	if i>len then return false -- invalid UTF-8 sequence
      if (bitAnd(ba[i],192)<>128) then return false -- invalid UTF-8 sequence
    end repeat
  end repeat
  
  return true -- no invalid UTF-8 sequence found
end

----------------------------------------
-- Converts bytearray to hex string (without spaces!)
-- @param {bytearray} ba
-- @return {string}
----------------------------------------
on toHexString (me, ba)
	str = ba.toHexString(1, ba.length)
	ret = ""
	cnt = str.word.count
	repeat with i = 1 to cnt
		put str.word[i] after ret
	end repeat
	return ret
end

----------------------------------------
-- Reads an IEEE 754 single-precision (32-bit) floating-point number from the byte stream
-- http://de.wikipedia.org/wiki/IEEE_754
-- @param {bytearray} ba
-- @return {float}
----------------------------------------
on readFloat32 (me, ba)

  if ba.endian=#littleendian then
    buf = bytearray()
    tmp = ba.readByteArray(4)
    repeat with i = 4 down to 1
      buf[i] = tmp[5-i]
    end repeat
  else
    buf = ba.readByteArray(4)
  end if
  
  c1 = buf[1]
  c2 = buf[2]
  
  neg = bitAnd(c1, 128)  
  exp = bitAnd(c1, 127) * 2 + bitAnd(c2, 128) / 128 - 127
  
  if exp = 128 then 
    return void
  else if exp = -127 then
    exp = -126
    mnt = 0
  else mnt = 1
  
  m = 0.5
  i = 64
  repeat with c = 1 to 3
    cn = buf[c+1]
    repeat while i
      if bitAnd(cn, i) then mnt = mnt + m
      m = m / 2
      i = i / 2
    end repeat
    i = 128
  end repeat
  
  position = position + 4
  if neg then return -mnt * power(2, exp)
  return mnt * power(2, exp)
  
end

----------------------------------------
-- Reads an IEEE 754 double-precision (64-bit) floating-point number from the byte stream
-- @param {bytearray} ba
-- @return {float}
----------------------------------------
on readFloat64 (me, ba)
    
  if ba.endian=#littleendian then
    buf = bytearray()
    tmp = ba.readByteArray(8)
    repeat with i = 8 down to 1
      buf[i] = tmp[9-i]
    end repeat
  else
    buf = ba.readByteArray(4)
  end if
  
  c1 = buf[1]
  c2 = buf[2]
  
  neg = bitAnd(c1, 128)
  exp = bitAnd(c1, 127) * 16 + bitAnd(c2, 240) / 16 - 1023
  
  if exp = 1024 then 
    return void
  else if exp = -1023 then
    exp = -1022
    mnt = 0
  else mnt = 1
  m = 0.5
  i = 8
  repeat with c = 1 to 7
    cn = buf[c+1]
    repeat while i
      if bitAnd(cn, i) then mnt = mnt + m
      m = m / 2
      i = i / 2
    end repeat
    i = 128
  end repeat
  
  position = position + 8
  if neg then return -mnt * power(2, exp)
  return mnt * power(2, exp)
end

----------------------------------------
-- Writes an IEEE 754 single-precision (32-bit) floating-point number to the byte stream.
-- @param {bytearray} ba
-- @param {float} floatVal
----------------------------------------
on writeFloat32 (me, ba, floatVal)
  floatVal = float(floatVal)
  if floatVal = 0 then
    ba.writeIn32(0)
    return
  end if
  if floatVal < 0 then
    floatVal = -floatVal
    neg = 128
  else neg = 0
  exp = 127
  if floatVal >= 1 then
    repeat while floatVal >= 2
      floatVal = floatVal / 2
      exp = exp + 1
    end repeat
  else
    repeat while floatVal < 1
      floatVal = floatVal * 2
      exp = exp - 1
    end repeat
  end if
  floatVal = floatVal - 1
  m = 0.5
  i = 64
  rslt = bytearray(4)
  rslt.writeInt8(bitAnd(exp, 254) / 2 + neg)
  val = bitAnd(exp, 1) * 128  
  repeat with c = 1 to 3
    repeat while i
      if floatVal >= m then
        floatVal = floatVal - m
        val = val + i
      end if
      m = m / 2.0
      i = i / 2
    end repeat
    rslt.writeInt8(val)
    i = 128
    val = 0
  end repeat  
  if ba.endian=#littleendian then
    tmp = byterray(4)
    repeat with i = 1 to 4
      tmp[i] = rslt[5-i]
    end repeat
    ba.writeByteArray(tmp)
  else
    ba.writeByteArray(rslt)
  end if
end

----------------------------------------
-- Writes an IEEE 754 double-precision (64-bit) floating-point number to the byte stream.
-- @param {bytearray} ba
-- @param {float} floatVal
----------------------------------------
on writeFloat64 (me, ba, floatVal)
  floatVal = float(floatVal)
  if floatVal = 0 then
         ba.writeIn32(0)
      ba.writeIn32(0)
    return
  end if  
  if floatVal < 0 then
    floatVal = -floatVal
    neg = 128
  else neg = 0
  exp = 1023
  if floatVal >= 1 then
    repeat while floatVal >= 2
      floatVal = floatVal / 2
      exp = exp + 1
    end repeat
  else
    repeat while floatVal < 1
      floatVal = floatVal * 2
      exp = exp - 1
    end repeat
  end if
  floatVal = floatVal - 1
  m = 0.5
  i = 8
  rslt = bytearray(8)
  rslt.writeInt8(bitAnd(exp, 2032) / 16 + neg)
  val = bitAnd(exp, 15) * 16
  repeat with c = 1 to 7
    repeat while i
      if floatVal >= m then
        floatVal = floatVal - m
        val = val + i
      end if
      m = m / 2
      i = i / 2
    end repeat
    rslt.writeInt8(val)
    i = 128
    val = 0
  end repeat  
  if ba.endian=#littleendian then
     tmp = byterray(8)
    repeat with i = 1 to 8
      tmp[i] = rslt[9-i]
    end repeat
    ba.writeByteArray(tmp)
  else
    ba.writeByteArray(rslt)
  end if
end
